# coding: utf-8
u"""Валидаторы для простых (текстовых, числовых и т.п.) полей модели Django."""
from functools import partial
from itertools import cycle
from itertools import izip
import re

from django.core.exceptions import ValidationError

from . import validate_value


# =============================================================================
# СНИЛС
# =============================================================================
_snils_re = re.compile(r'^\d{3}-\d{3}-\d{3} \d{2}$')  # СНИЛС


def snils_validator(value):
    u"""Валидатор СНИЛС.

    Проверяет корректность формата и контрольное число (последние две цифры).

    :raises django.core.exceptions.ValidationError: Если аргумент value
        содержит значение, несоответствующее формату СНИЛС, либо контрольное
        число некорректно.
    """
    value = unicode(value)

    if not _snils_re.match(value):
        raise ValidationError(u'СНИЛС должен быть в формате ###-###-### ##')

    if value[:11] <= u'001-001-998':
        return

    numbers = (int(ch) for ch in reversed(value[:11]) if ch.isdigit())
    checksum = int(value[12:14])
    summa = sum(i * n for i, n in enumerate(numbers, 1))
    if summa > 101:
        summa %= 101

    if (summa < 100 and summa != checksum or
            summa in (100, 101) and checksum != 0):
        raise ValidationError(u'не пройдена проверка контрольного числа')


is_snils_valid = partial(validate_value, validator=snils_validator)


# =============================================================================
# ИНН
# =============================================================================


def _check_inn_checksum(coefficients, numbers, checksum):
    summa = sum(c * n for c, n in izip(coefficients, numbers))
    if summa % 11 % 10 != checksum:
        raise ValidationError(u'Не пройдена проверка контрольного числа')


def inn10_validator(value):
    u"""Валидатор для ИНН юридического лица (10 цифр).

    Проверяет корректность формата и контрольное число (последняя цифра).

    :raises django.core.exceptions.ValidationError: Если аргумент value
        содержит значение, несоответствующее формату ИНН, либо не пройдена
        проверка контрольного числа.
    """
    value = unicode(value)

    if len(value) != 10 or not value.isdigit():
        raise ValidationError(u'ИНН должен быть 10-тизначным числом')

    _check_inn_checksum((2, 4, 10, 3, 5, 9, 4, 6, 8),
                        (int(ch) for ch in value[:-1]),
                        int(value[-1]))


def inn12_validator(value):
    u"""Валидатор для ИНН физического лица (12 цифр).

    Проверяет корректность формата и контрольное число (последние 2 цифры).

    :raises django.core.exceptions.ValidationError: Если аргумент value
        содержит значение, несоответствующее формату ИНН, либо не пройдена
        проверка контрольного числа.
    """
    value = unicode(value)

    if len(value) != 12 or not value.isdigit():
        raise ValidationError(u'ИНН должен быть 12-тизначным числом')

    _check_inn_checksum((7, 2, 4, 10, 3, 5, 9, 4, 6, 8),
                        (int(ch) for ch in value[:-2]),
                        int(value[-2]))

    _check_inn_checksum((3, 7, 2, 4, 10, 3, 5, 9, 4, 6, 8),
                        (int(ch) for ch in value[:-1]),
                        int(value[-1]))


def inn_validator(value):
    u"""Валидатор для поля ИНН.

    Проверяет корректность формата и контрольное число (1 или 2 последние
    цифры).

    :raises django.core.exceptions.ValidationError: Если аргумент value
        содержит значение, несоответствующее формату ИНН, либо не пройдена
        проверка контрольного числа.
    """
    value = unicode(value)

    if len(value) not in (10, 12) or not value.isdigit():
        raise ValidationError(u'ИНН должен быть 10-ти или 12-тизначным числом')

    if len(value) == 10:
        inn10_validator(value)
    else:
        inn12_validator(value)


is_inn_valid = partial(validate_value, validator=inn_validator)
is_inn10_valid = partial(validate_value, validator=inn10_validator)
is_inn12_valid = partial(validate_value, validator=inn12_validator)


# =============================================================================
# КПП, ОКАТО, ОКТМО, ОКПО, ОГРН, ОКВЭД, ОКОПФ, ОКФС
# =============================================================================


def kpp_validator(value):
    u"""Валидатор для КПП (кода причины постановки на налоговый учет)."""
    value = unicode(value)

    if len(value) != 9 or not value.isdigit():
        raise ValidationError(u'КПП должен быть 9-тизначным числом')


is_kpp_valid = partial(validate_value, validator=kpp_validator)


def checksum_pr_50_1_024(value):
    u"""Расчет контрольного числа в соответствии с ПР 50.1.024-2005.

    Контрольное числорассчитывается следующим образом:
        1. Разрядам кода в общероссийском классификаторе, начиная со старшего
           разряда,присваивается набор весов, соответствующий натуральному
           ряду чисел от 1 до 10. Если разрядность кода больше 10, то набор
           весов повторяется.
        2. Каждая цифра кода умножается на вес разряда и вычисляется сумма
           полученных произведений.
        3. Контрольное число для кода представляет собой остаток от деления
           полученной суммы на модуль 11.
        4. Контрольное число должно иметь один разряд, значение которого
           находится в пределах от 0 до 9.

    Если получается остаток, равный 10, то для обеспечения одноразрядного
    контрольного числа необходимо провести повторный расчет, применяя вторую
    последовательность весов, сдвинутую на два разряда влево (3, 4, 5,...).

    Если в случае повторного расчета остаток от деления вновь сохраняется
    равным 10, то значение контрольного числа проставляется равным 0.
    """
    summa = sum(
        int(digit) * weight
        for digit, weight in izip(value, cycle(range(1, 11)))
    )

    remainder = summa % 11
    if remainder < 10:
        result = remainder
    else:
        summa = sum(
            int(digit) * weight
            for digit, weight in izip(value, cycle(range(3, 13)))
        )
        remainder = summa % 11
        if remainder < 10:
            result = remainder
        else:
            result = 0

    return result


def okato_validator(value):
    u"""Валидатор для ОКАТО.

    ОКАТО - общероссийский классификатор объектов административно-
    территориального деления.

    .. seealso::
       `Контрольное число ОКАТО <http://kontragent.info/articles/okato>`_
    """
    value = unicode(value)

    if not value.isdigit():
        raise ValidationError(u'ОКАТО должен состоять только из цифр')

    value_length = len(value)
    if value_length in (3, 6, 9, 12):
        # В последнем разряде указано контрольное число, проверим его
        if not checksum_pr_50_1_024(value[:-1]) != int(value[-1]):
            raise ValidationError(u'Не пройдена проверка контрольного числа')

    elif value_length not in (2, 5, 8, 11):
        raise ValidationError(
            u'Код ОКАТО должен состоять из 2, 5, 8 или 11 цифр'
        )


is_okato_valid = partial(validate_value, validator=okato_validator)


def oktmo_validator(value):
    u"""Валидатор для ОКТМО.

    ОКТМО - Общероссийский классификатор территорий муниципальных образований.
    """
    value = unicode(value)

    if not value.isdigit():
        raise ValidationError(u'ОКАТО должен состоять только из цифр')

    value_length = len(value)
    if value_length not in (2, 5, 8, 11):
        raise ValidationError(
            u'ОКАТО не может состоять из {} цифр'.format(value_length)
        )


is_oktmo_valid = partial(validate_value, validator=oktmo_validator)


def okpo_validator(value):
    u"""Валидатор кода ОКПО.

    ОКПО - Общероссийский классификатор предприятий и организаций.
    """
    value = unicode(value)

    if not value.isdigit():
        raise ValidationError(u'ОКПО должен состоять только из цифр')

    value_length = len(value)
    if value_length not in (8, 10):
        raise ValidationError(u'ОКПО должен состоять из 8-ми или 10-ти цифр')


is_okpo_valid = partial(validate_value, validator=okpo_validator)


def _check_ogrn_checksum(value):
    u"""Проверка контрольного числа ОГРН(ИП).

    Младший разряд остатка от деления предыдущего (n-1)-значного числа на (n-2)
    должен быть равен n-ому знаку ОГРН, где n – длина ОГРН. Если остаток от
    деления равен 10, то контрольное число равно 0 (нулю).
    """
    value_length = len(value)  # количество цифр в ОГРН
    divisor = value_length-2  # делитель для вычисления контрольного числа
    number = int(value[:value_length-1])  # ОГРН
    control_number = int(value[-1])  # контрольное число

    if number % divisor % 10 != control_number:
        raise ValidationError(u'Не пройдена проверка контрольного числа')


def _ogrn_validator(value, valid_length, title):
    value = unicode(value)

    if not value.isdigit():
        raise ValidationError(u'ОГРН должен состоять только из цифр')
    if len(value) not in valid_length:
        raise ValidationError(
            u'{} должен состоять из {} цифр!'.format(
                title,
                u' или '.join(unicode(l) for l in valid_length)
            )
        )

    _check_ogrn_checksum(value)


def ogrn13_validator(value):
    u"""Валитор ОГРН.

    ОГРН - Основной государственный регистрационный номер.
    """
    _ogrn_validator(value, (13,), u'ОГРН')


def ogrn15_validator(value):
    u"""Валидатор ОГРНИП.

    ОГРНИП - Основной государственный регистрационный номер индивидуального
    предпринимателя.
    """
    _ogrn_validator(value, (15,), u'ОГРНИП')


def ogrn_validator(value):
    u"""Валидатор ОГРН(ИП).

    ОГРН(ИП) - Основной государственный регистрационный номер (индивидуального
    предпринимателя).
    """
    _ogrn_validator(value, (13, 15), u'ОГРН(ИП)')


is_ogrn_valid = partial(validate_value, validator=ogrn_validator)
is_ogrn13_valid = partial(validate_value, validator=ogrn13_validator)
is_ogrn15_valid = partial(validate_value, validator=ogrn15_validator)


okved_re = re.compile(r'^(\d{2}|\d{2}\.\d{1,2}|\d{2}\.\d{1,2}\.\d{1,2})$')


def okved_validator(value):
    u"""Валидатор кода ОКВЭД.

    ОКВЭД - Общероссийский классификатор видов экономической деятельности.
    """
    value = unicode(value)

    if not value.isdigit():
        raise ValidationError(u'ОКВЭД должен состоять только из цифр')

    value_length = len(value)
    if value_length < 2 or 6 < value_length:
        raise ValidationError(
            u'Код ОКВЭД не может состоять из {} цифр'.format(value_length)
        )


is_okved_valid = partial(validate_value, validator=okved_validator)


okopf_re = re.compile(r'^(\d \d\d \d\d)$')


def okopf_validator(value):
    u"""Валидатор кода ОКОПФ.

    ОКОПФ - Общероссийский классификатор организационно-правовых форм.
    """
    value = unicode(value)

    if not value.isdigit():
        raise ValidationError(u'ОКОПФ должен состоять только из цифр')

    value_length = len(value)
    if value_length != 5:
        raise ValidationError(u'Код ОКОПФ должен состоять из 5-ти цифр')


is_okopf_valid = partial(validate_value, validator=okopf_validator)


def okfs_validator(value):
    u"""Валидатор кода ОКФС.

    ОКФС - Общероссийский классификатор форм собственности.
    """
    value = unicode(value)

    if not value.isdigit():
        raise ValidationError(u'ОКФС должен состоять только из цифр')

    value_length = len(value)
    if value_length != 2:
        raise ValidationError(u'Код ОКФС должен состоять из 2-ти цифр')


is_okfs_valid = partial(validate_value, validator=okfs_validator)
